/*
  Copyright Red Hat, Inc. 2008

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; version 2.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to the
  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
  MA 02139, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <limits.h>
#include <sys/stat.h>
#include <errno.h>
#include <libgen.h>
#include <string.h>
#include "readlinkr.h"


static int
parse_relative(char *path, char *buf, size_t buflen)
{
	char *start = buf;
	char mod_path[PATH_MAX], tmp[PATH_MAX];
	char *dn_ptr;

	//printf("%s ... %s\n", path, buf);

	memset(tmp, 0, buflen);

	strncpy(mod_path, path, buflen);
	memcpy(tmp, mod_path, buflen);
	dn_ptr = dirname(mod_path); /* Strip off the original path */
	strncpy(mod_path, dn_ptr, buflen);


	while(1) {
		//printf("mod_path = %s | start = %s\n", mod_path, start);
		if (strncmp(start, "./", 2) == 0) {
			start += 2;
			continue;
		}

		if (strncmp(start, "../", 3) == 0) {
			start += 3;

			strncpy(tmp, mod_path, buflen);
			dn_ptr = dirname(tmp);
			if (*dn_ptr == '.') {
				/* file name - handle how?! */
				free(tmp);
				free(mod_path);
				errno = EINVAL;
				return -1;
			}

			/* parent directory */
			strncpy(mod_path, dn_ptr, buflen);
			continue;
		}

		break;
	}

	snprintf(tmp, buflen, "%s/%s", mod_path, start);
	memcpy(buf, tmp, buflen);

	return 0;
}


int
readlink_recurse(char *path, char *buf, size_t buflen)
{
	char real[PATH_MAX];
	char tmp[PATH_MAX];
	struct stat st_buf;

	memset(buf, 0, buflen);
	snprintf(real, sizeof(real), "%s", path);

	do {
		if (lstat(real, &st_buf) < 0) {
			return -1;
		}

		if (!S_ISLNK(st_buf.st_mode)) {
			/* not a link! */
			break;
		}

		memset(tmp, 0, sizeof(tmp));
		if (readlink(real, tmp, sizeof(tmp)-1) < 0) {
			return -1;
		}

		memcpy(real, tmp, sizeof(tmp));

		if (real[0]=='/') {
			continue;
		}

		/* todo - handle relative links */
		if (parse_relative(path, real, sizeof(real)) < 0) {
			return -1;
		}

		//printf("real = %s\n", real);

	} while (1);

	snprintf(buf, buflen, "%s", real);

	return 0;
}


#ifdef STANDALONE
int
main(int argc, char **argv)
{
	char buf[PATH_MAX];

	if (argc < 2) {
		printf("usage: %s path\n", argv[0]);
		return -1;
	}

	if (readlink_recurse(argv[1], buf, sizeof(buf)) < 0) {
		perror("readlink_recurse");
		return -1;
	}

	printf("%s -> %s\n", argv[1], buf);

	return 0;
}

#endif

